/* lsmbox -- list nr of total and read messages for one or several mailboxes
 *
 * Copyright (C) 2002 David Weinehall
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#define _GNU_SOURCE
#include <pwd.h>
#include <glib.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <curses.h>
#include <getopt.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>

#if __GNUC__ == 2 && __GNUC_MINOR__ < 96
#define __builtin_expect(x, expected_value) (x)
#endif

#ifndef likely
#define likely(x)       __builtin_expect((x), 1)
#endif /* likely */

#ifndef unlikely
#define unlikely(x)     __builtin_expect((x), 0)
#endif /* unlikely */

#define PRG_NAME		"lsmbox"
#define PRG_VERSION		"v1.2.2"

#define SPOOLFILE_PATH		"/var/mail"	/* username follows */
#define MAILBOX_FOLDER_PATH	"Mail"		/* $HOME preceeds */

#define SET_STR			"set"
#define UNSET_STR		"unset"
#define MARK_OLD_STR		"mark_old"
#define SPOOLFILE_STR		"spoolfile"
#define MAILBOX_FOLDER_STR	"folder"
#define MAILBOX_STR		"mailboxes"

#define FROM_STR		"From "
#define STATUS_STR		"Status: "
#define CONTENT_LENGTH_STR	"Content-Length: "

#define BUF_SIZE		1024

#define B_PRINT(__str...)					\
	do {							\
		if (continuous) {				\
			attron(A_BOLD);				\
			printw(__str);				\
			attroff(A_BOLD);			\
			g_assert(!refresh());			\
		} else {					\
			g_print(__str);				\
		}						\
	} while (0)

#define R_PRINT(__str...)					\
	do {							\
		if (continuous) {				\
			printw(__str);				\
			g_assert(!refresh());			\
		} else {					\
			g_print(__str);				\
		}						\
	} while (0)

enum error_t {
	E_OK,
	E_SYNTAX,
	E_UNKNOWN
};

gchar buf[BUF_SIZE];

gchar *homepath = NULL;
gchar *spoolfilepath = NULL;
gchar *mailfolderpath = NULL;
gchar **mailboxes = NULL;
gboolean markold = TRUE;
gboolean continuous = FALSE;
gboolean quit = FALSE;

struct mailstat {
	gchar *fname;
	gulong total;
	gulong new;
	gulong old;
	off_t size;
	time_t mtime;
};

struct mailstat *mboxes = NULL;

/**
 * Show usage-information for the program
 *
 * @version v1.00
 */
void usage(void)
{
	g_print("Usage: %s [OPTION]... [MAILBOX]...\n"
		"List the number of total, old unread, and new messages\n"
		"for the specified mailbox(es).\n"
		"\n"
		"  -c, --continuous  update statistics continuously\n"
		"      --help        display this help and exit\n"
		"      --version     output version information and exit\n"
		"\n"
		"Report bugs to <tao@kernel.org>\n", PRG_NAME);
}

/**
 * Show version-information for the program
 *
 * @version v1.00
 */
void version(void)
{
	g_print("%s %s\n"
		"Written by David Weinehall.\n"
		"\n"
		"Copyright (C) 2002 David Weinehall (tao@kernel.org)\n"
		"This is free software; see the source for copying "
		"conditions.  There is NO\n"
		"warranty; not even for MERCHANTABILITY or FITNESS FOR "
		"A PARTICULAR PURPOSE.\n",
		PRG_NAME, PRG_VERSION);
}

/**
 * Reset curses
 *
 * @version v1.00
 */
void reset_curses(void)
{
	g_assert(!endwin());
}

/**
 * Setup curses
 *
 * @version v1.01
 *  @return TRUE on success, FALSE on failure
 */
gboolean init_curses(void)
{
	initscr();
	g_atexit(reset_curses);

	if (cbreak() == ERR || noecho() == ERR || nonl() == ERR ||
	    curs_set(0) == ERR || nodelay(stdscr, TRUE) == ERR ||
	    intrflush(stdscr, FALSE) == ERR)
		return FALSE;

	g_assert(!refresh());

	return TRUE;
}

/**
 * Skip all leading whitespace in a string
 *
 * @version v1.00
 *
 *   @param str The string to process
 *  @return A new pointer into the string with leading whitespace skipped
 */
gchar *skip_whitespace(gchar *str)
{
	for (; *str != 0; str++)
		if (!isspace(*str))
			break;
	return str;
}

/**
 * Skip past a configuration-option to its value
 *
 * @version v1.00
 *
 *   @param str The string to process
 *  @return A new pointer to the value of the configuration-option
 */
gchar *skip_config_option(gchar *str)
{
	for (; *str != 0; str++)
		if (isspace(*str) || *str == '=')
			break;

	for (; *str != 0; str++)
		if (!isspace(*str) && *str != '=')
			break;

	return str;
}

/**
 * Expand a path-name
 *
 * @version v1.00
 *
 *   @param str The path to expand
 *  @return A pointer to the string with the new path;
 *          needs to be freed after use
 */
gchar *path_expand(const gchar *str)
{
	gchar *tmp = NULL, *tmp2 = NULL;
	struct passwd *pw;

	if (!strcmp(str, "!"))
		tmp = g_strdup(spoolfilepath);
	else if (strchr(str, '/'))
		tmp = g_strdup(str);
	else if (*str == '+' || *str == '=')
		tmp = g_strjoin("/", mailfolderpath, str + 1, NULL);
	else
		tmp = g_strjoin("/", mailfolderpath, str, NULL);

	if (*tmp == '~') {
		gchar *rest = NULL;

		if (*(tmp + 1) == '/') {
			g_assert((pw = getpwuid(getuid())));
		} else {
			gchar *name = g_strdup(tmp + 1);
			gchar *end;
			end = strchr(name, '/');
			*end = '\0';
			g_assert((pw = getpwnam(name)));
			g_free(name);
		}
		rest = strchr(tmp, '/') + 1;
		tmp2 = g_strjoin("/", pw->pw_dir, rest, NULL);
	} else {
		tmp2 = g_strdup(tmp);
	}

	g_free(tmp);

	return tmp2;
}

/**
 * Parse the configuration-file
 *
 * @version v1.07
 *
 *   @param fname The name of the configuration-file to parse
 *  @return TRUE on success, FALSE on failure
 */
gboolean parse_config_file(const gchar *fname)
{
	FILE *fp;
	gchar *str;
	gchar *mb = NULL;
	gint line = 0;
	error_t error;

	if (!(fp = fopen(fname, "r"))) {
		if (errno == ENOENT)
			errno = 0;
		return !errno;
	}

	/* Poison */
	buf[BUF_SIZE - 1] = (gchar)0xff;

	while ((str = fgets(buf, BUF_SIZE, fp))) {
		/* We got a single line larger than 64kB */
		if (buf[BUF_SIZE - 1] != (gchar)0xff) {
			errno = EFBIG;
			return FALSE;
		}

		error = E_OK;

		line++;

		/* Ignore all comments */
		if (*str == '#')
			continue;

		str = skip_whitespace(str);

		/* Skip empty lines */
		if (!*str)
			continue;

		/* Remove the trailing newline */
		*(str + strlen(str) - 1) = '\0';

		if (!strncmp(str, SET_STR, strlen(SET_STR))) {
			str = skip_config_option(str);

			if (!strncmp(str, SPOOLFILE_STR,
				     strlen(SPOOLFILE_STR))) {
				str = skip_config_option(str);

				if (!str) {
					error = E_SYNTAX;
				} else {
					g_free(spoolfilepath);
					spoolfilepath = g_strdup(str);
				}
			} else if (!strncmp(str, MAILBOX_FOLDER_STR,
					    strlen(MAILBOX_FOLDER_STR))) {
				str = skip_config_option(str);

				if (!str) {
					error = E_SYNTAX;
				} else {
					g_free(mailfolderpath);
					mailfolderpath = g_strdup(str);
				}
			} else if (!strncmp(str, MARK_OLD_STR,
					    strlen(MARK_OLD_STR))) {
				markold = TRUE;
			} else {
				error = E_UNKNOWN;
			}
		} else if (!strncmp(str, UNSET_STR, strlen(UNSET_STR))) {
			str = skip_config_option(str);

			if (!strncmp(str, SPOOLFILE_STR,
				     strlen(SPOOLFILE_STR))) {
				g_free(spoolfilepath);
				spoolfilepath = NULL;
			} else if (!strncmp(str, MAILBOX_FOLDER_STR,
					    strlen(MAILBOX_FOLDER_STR))) {
				g_free(mailfolderpath);
				mailfolderpath = NULL;
			} else if (!strncmp(str, MARK_OLD_STR,
				strlen(MARK_OLD_STR))) {
				markold = FALSE;
			} else {
				error = E_UNKNOWN;
			}
		} else if (!strncmp(str, MAILBOX_STR, strlen(MAILBOX_STR))) {
			gchar *tmp;

			str = skip_config_option(str);

			if (!str) {
				error = E_SYNTAX;
			} else {
				if (mb)
					tmp = g_strjoin(" ", mb, str, NULL);
				else
					tmp = g_strdup(str);

				g_free(mb);
				mb = tmp;
			}
		} else {
			error = E_SYNTAX;
		}

		switch (error) {
		case E_UNKNOWN:
			g_print("Unknown option on line: %d inside: %s;"
				"ignoring.\n",
				line, fname);
			break;
		case E_SYNTAX:
			g_print("Syntax error on line: %d inside: %s;"
				"ignoring.\n",
				line, fname);
			break;
		case E_OK:
		default:
			break;
		}
	}

	if (mb)
		mailboxes = g_strsplit(mb, " ", 0);

	g_free(mb);

	return !(fclose(fp) == EOF);
}

/**
 * Parse a mailbox in mbox/MMDF format
 *
 * @version v1.12
 *
 *   @param mbox A mailstat structure for the mailbox
 *  @return TRUE on success, FALSE on failure
 */
gboolean parse_mailbox(struct mailstat *mbox)
{

	FILE *fp;
	gchar *str;
	gulong clen = 0;
	gboolean msg = FALSE;
	gboolean old = FALSE;
	gboolean read = TRUE;
	gboolean msgbody = FALSE;

	mbox->total = 0;
	mbox->new = 0;
	mbox->old = 0;

	if (!(fp = fopen(mbox->fname, "r")))
		return FALSE;

	/* Poison */
	buf[BUF_SIZE - 1] = (gchar)0xff;

	while (likely(str = fgets(buf, BUF_SIZE, fp))) {
		/* We got a single line larger than 64kB */
		if (unlikely(buf[BUF_SIZE - 1] != (gchar)0xff)) {
			errno = EFBIG;
			goto EXIT;
		}

		if (unlikely(!strncmp(str, FROM_STR,
				      strlen(FROM_STR)))) {
			msg = TRUE;
			mbox->total++;

			if (old)
				mbox->old++;
			else if (!read)
				mbox->new++;

			old = FALSE;
			read = FALSE;
			msgbody = FALSE;

			if (continuous && unlikely(getch() == 'q')) {
				quit = TRUE;
				goto EXIT;
			}
		} else if (likely(msgbody)) {
			continue;
		} else if (unlikely(!strncmp(str, STATUS_STR,
					     strlen(STATUS_STR)))) {
			if (str[8] == 'R')
				read = TRUE;
			else if (str[8] == 'O')
				old = TRUE;
		} else if (unlikely(!strncmp(str, CONTENT_LENGTH_STR,
				             strlen(CONTENT_LENGTH_STR)))) {
			clen = strtoul(str + strlen(CONTENT_LENGTH_STR),
				       NULL, 10);

		        if (clen == ULONG_MAX && errno == ERANGE) {
		                errno = ERANGE;
				goto EXIT;
			} else {
				if (fseek(fp, clen, SEEK_CUR))
					goto EXIT;
				clen = 0;
			}
		} else if (*str == '\n') {
			msgbody = TRUE;
		}
	}

	if (old)
		mbox->old++;
	else if (!read)
		mbox->new++;

EXIT:
	return !(fclose(fp) == EOF);
}

/**
 * Show statistics for all mailboxes in the mboxes array
 *
 * @version v1.04
 *
 *  @return 0 on success, errno on failure
 */
gint show_mailbox_stats(void)
{
	gint i;
	struct mailstat totals;

	B_PRINT("mailbox:                                     "
		"total:      %s:%s\n",
		markold ? "new" : "unread",
		markold ? "        old:" : "");

	for (i = 0; mboxes[i].fname; i++) {
		struct stat st;

		if (stat(mboxes[i].fname, &st)) {
			if (errno == ENOENT) {
				errno = 0;
				continue;
			} else {
				goto EXIT;
			}
		}

		if (st.st_mtime <= mboxes[i].mtime &&
		    st.st_size == mboxes[i].size)
			continue;

		if ((!parse_mailbox(&mboxes[i]) && errno != ENOENT) || quit)
			goto EXIT;

		if (errno == ENOENT) {
			errno = 0;
			continue;
		} else {
			if (!markold)
				R_PRINT("%-43s  %-10lu  %-10lu\n",
					mboxes[i].fname, mboxes[i].total,
					mboxes[i].new + mboxes[i].old);
			else
				R_PRINT("%-43s  %-10lu  %-10lu  %-10lu\n",
					mboxes[i].fname, mboxes[i].total,
					mboxes[i].new, mboxes[i].old);
		}
	}

	totals.total = 0;
	totals.new = 0;
	totals.old = 0;

	for (i = 0; mboxes[i].fname; i++) {
		totals.total += mboxes[i].total;
		totals.new += mboxes[i].new;
		totals.old += mboxes[i].old;
	}

	R_PRINT("\nSummary:                                     %-10lu",
		totals.total);

	if (!markold)
		R_PRINT("  %-10lu\n",
			totals.new + totals.old);
	else
		R_PRINT("  %-10lu  %-10lu\n",
			totals.new, totals.old);

EXIT:
	if (continuous)
		move(0, 0);

	return errno;
}

/**
 * The program's main-function
 *
 * @version v1.12
 *
 *   @param argc The number of arguments
 *   @param argv The arguments
 *  @return 0 on success, errno on failure
 */
int main(int argc, char *argv[])
{
	gint i;
	gint optc;
	gint opt_index;
	gint start, end;
	extern gint optind;
	extern gchar *optarg;

	gchar **list = NULL;
	gchar *userrcpath = NULL;

	struct option const options[] = {
		{ "continuous", no_argument, 0, 'c' },
		{ "help", no_argument, 0, 'h' },
		{ "version", no_argument, 0, 'V' },
		{ 0, 0, 0, 0 }
	};

	errno = 0;

	while ((optc = getopt_long(argc, argv, "c",
				   options, &opt_index)) != -1) {
		switch (optc) {
		case 'c':
			continuous = TRUE;
			break;
		case 'h':
			usage();
			goto EXIT;
		case 'V':
			version();
			goto EXIT;
		default:
			usage();
			errno = EINVAL;
			goto EXIT;
		}
	}

	g_assert(homepath = g_strdup(g_get_home_dir()));

	userrcpath = g_strjoin(NULL, homepath, "/.lsmboxrc", NULL);

	if (!parse_config_file(userrcpath))
		goto EXIT;

	if (!spoolfilepath) {
		spoolfilepath = g_strdup(SPOOLFILE_PATH);
		spoolfilepath = g_strjoin("/", SPOOLFILE_PATH,
					  g_get_user_name(), NULL);
	}

	if (!mailfolderpath)
		mailfolderpath = g_strjoin("/", homepath,
					   MAILBOX_FOLDER_PATH, NULL);

	if ((argv[optind] == NULL || (argc < 2)) && !mailboxes) {
		usage();
		errno = EINVAL;
		goto EXIT;
	}

	if (argv[optind]) {
		start = optind;
		end = argc;
		list = argv;
	} else {
		start = 0;
		for (end = 0; mailboxes[end]; end++)
			;
		list = mailboxes;
	}

	if (continuous)
		if (!init_curses())
			goto EXIT;

	mboxes = g_new0(struct mailstat, end - start + 1);

	for (i = start; i < end; i++)
		mboxes[i - start].fname = path_expand(list[i]);

	do {
		show_mailbox_stats();
	} while (!errno && continuous && !quit);

EXIT:
	for (i = 0; mboxes && mboxes[i].fname; i++)
		g_free(mboxes[i].fname);

	g_free(mboxes);
	g_free(spoolfilepath);
	g_free(userrcpath);
	g_free(mailfolderpath);
	g_strfreev(mailboxes);

	return errno;
}
